@@ -3,25 +3,40 @@ package org.jsoup;
 import org.jsoup.nodes.Document;
 import org.jsoup.parser.Parser;
 
+import javax.annotation.Nullable;
+import javax.net.ssl.SSLSocketFactory;
+import java.io.BufferedInputStream;
 import java.io.IOException;
 import java.io.InputStream;
+import java.net.CookieStore;
 import java.net.Proxy;
 import java.net.URL;
 import java.util.Collection;
+import java.util.List;
 import java.util.Map;
 
 /**
- * A Connection provides a convenient interface to fetch content from the web, and parse them into Documents.
- * <p>
- * To get a new Connection, use {@link org.jsoup.Jsoup#connect(String)}. Connections contain {@link Connection.Request}
- * and {@link Connection.Response} objects. The request objects are reusable as prototype requests.
- * </p>
- * <p>
- * Request configuration can be made using either the shortcut methods in Connection (e.g. {@link #userAgent(String)}),
- * or by methods in the Connection.Request object directly. All request configuration must be made before the request is
- * executed.
- * </p>
+ The Connection interface is a convenient HTTP client and session object to fetch content from the web, and parse them
+ into Documents.
+ <p>To start a new session, use either {@link org.jsoup.Jsoup#newSession()} or {@link org.jsoup.Jsoup#connect(String)}.
+ Connections contain {@link Connection.Request} and {@link Connection.Response} objects (once executed). Configuration
+ settings (URL, timeout, useragent, etc) set on a session will be applied by default to each subsequent request.</p>
+ <p>To start a new request from the session, use {@link #newRequest()}.</p>
+ <p>Cookies are stored in memory for the duration of the session. For that reason, do not use one single session for all
+ requests in a long-lived application, or you are likely to run out of memory, unless care is taken to clean up the
+ cookie store. The cookie store for the session is available via {@link #cookieStore()}. You may provide your own
+ implementation via {@link #cookieStore(java.net.CookieStore)} before making requests.</p>
+ <p>Request configuration can be made using either the shortcut methods in Connection (e.g. {@link #userAgent(String)}),
+ or by methods in the Connection.Request object directly. All request configuration must be made before the request is
+ executed. When used as an ongoing session, initialize all defaults prior to making multi-threaded {@link
+#newRequest()}s.</p>
+ <p>Note that the term "Connection" used here does not mean that a long-lived connection is held against a server for
+ the lifetime of the Connection object. A socket connection is only made at the point of request execution ({@link
+#execute()}, {@link #get()}, or {@link #post()}), and the server's response consumed.</p>
+ <p>For multi-threaded implementations, it is important to use a {@link #newRequest()} for each request. The session may
+ be shared across threads but a given request, not.</p>
  */
+@SuppressWarnings("unused")
 public interface Connection {
 
     /**
@@ -45,6 +60,13 @@ public interface Connection {
         }
     }
 
+    /**
+     Creates a new request, using this Connection as the session-state and to initialize the connection settings (which may then be independently on the returned Connection.Request object).
+     @return a new Connection object, with a shared Cookie Store and initialized settings from this Connection and Request
+     @since 1.14.1
+     */
+    Connection newRequest();
+
     /**
      * Set the request URL to fetch. The protocol must be HTTP or HTTPS.
      * @param url URL to connect to
@@ -60,11 +82,11 @@ public interface Connection {
     Connection url(String url);
 
     /**
-     * Set the proxy to use for this request. Set to <code>null</code> to disable.
+     * Set the proxy to use for this request. Set to <code>null</code> to disable a previously set proxy.
      * @param proxy proxy to use
      * @return this Connection, for chaining
      */
-    Connection proxy(Proxy proxy);
+    Connection proxy(@Nullable Proxy proxy);
 
     /**
      * Set the HTTP proxy to use for this request.
@@ -83,17 +105,22 @@ public interface Connection {
     Connection userAgent(String userAgent);
 
     /**
-     * Set the request timeouts (connect and read). If a timeout occurs, an IOException will be thrown. The default
-     * timeout is <b<30 seconds</b> (30000 millis). A timeout of zero is treated as an infinite timeout.
+     * Set the total request timeout duration. If a timeout occurs, an {@link java.net.SocketTimeoutException} will be thrown.
+     * <p>The default timeout is <b>30 seconds</b> (30,000 millis). A timeout of zero is treated as an infinite timeout.
+     * <p>Note that this timeout specifies the combined maximum duration of the connection time and the time to read
+     * the full response.
      * @param millis number of milliseconds (thousandths of a second) before timing out connects or reads.
      * @return this Connection, for chaining
+     * @see #maxBodySize(int)
      */
     Connection timeout(int millis);
 
     /**
      * Set the maximum bytes to read from the (uncompressed) connection into the body, before the connection is closed,
-     * and the input truncated. The default maximum is 1MB. A max size of zero is treated as an infinite amount (bounded
-     * only by your patience and the memory available on your machine).
+     * and the input truncated (i.e. the body content will be trimmed). <b>The default maximum is 2MB</b>. A max size of
+     * <code>0</code> is treated as an infinite amount (bounded only by your patience and the memory available on your
+     * machine).
+     *
      * @param bytes number of bytes to read from the input before truncating
      * @return this Connection, for chaining
      */
@@ -140,23 +167,11 @@ public interface Connection {
     Connection ignoreContentType(boolean ignoreContentType);
 
     /**
-     * Disable/enable TLS certificates validation for HTTPS requests.
-     * <p>
-     * By default this is <b>true</b>; all
-     * connections over HTTPS perform normal validation of certificates, and will abort requests if the provided
-     * certificate does not validate.
-     * </p>
-     * <p>
-     * Some servers use expired, self-generated certificates; or your JDK may not
-     * support SNI hosts. In which case, you may want to enable this setting.
-     * </p>
-     * <p>
-     * <b>Be careful</b> and understand why you need to disable these validations.
-     * </p>
-     * @param value if should validate TLS (SSL) certificates. <b>true</b> by default.
+     * Set custom SSL socket factory
+     * @param sslSocketFactory custom SSL socket factory
      * @return this Connection, for chaining
      */
-    Connection validateTLSCertificates(boolean value);
+    Connection sslSocketFactory(SSLSocketFactory sslSocketFactory);
 
     /**
      * Add a request data parameter. Request parameters are sent in the request query string for GETs, and in the
@@ -176,9 +191,23 @@ public interface Connection {
      * @param inputStream the input stream to upload, that you probably obtained from a {@link java.io.FileInputStream}.
      * You must close the InputStream in a {@code finally} block.
      * @return this Connections, for chaining
+     * @see #data(String, String, InputStream, String) if you want to set the uploaded file's mimetype.
      */
     Connection data(String key, String filename, InputStream inputStream);
 
+    /**
+     * Add an input stream as a request data parameter. For GETs, has no effect, but for POSTS this will upload the
+     * input stream.
+     * @param key data key (form item name)
+     * @param filename the name of the file to present to the remove server. Typically just the name, not path,
+     * component.
+     * @param inputStream the input stream to upload, that you probably obtained from a {@link java.io.FileInputStream}.
+     * @param contentType the Content Type (aka mimetype) to specify for this file.
+     * You must close the InputStream in a {@code finally} block.
+     * @return this Connections, for chaining
+     */
+    Connection data(String key, String filename, InputStream inputStream, String contentType);
+
     /**
      * Adds all of the supplied data to the request data parameters
      * @param data collection of data parameters
@@ -194,11 +223,15 @@ public interface Connection {
     Connection data(Map<String, String> data);
 
     /**
-     * Add a number of request data parameters. Multiple parameters may be set at once, e.g.: <code>.data("name",
-     * "jsoup", "language", "Java", "language", "English");</code> creates a query string like:
-     * <code>{@literal ?name=jsoup&language=Java&language=English}</code>
-     * @param keyvals a set of key value pairs.
-     * @return this Connection, for chaining
+     Add one or more request {@code key, val} data parameter pairs.<p>Multiple parameters may be set at once, e.g.:
+     <code>.data("name", "jsoup", "language", "Java", "language", "English");</code> creates a query string like:
+     <code>{@literal ?name=jsoup&language=Java&language=English}</code></p>
+     <p>For GET requests, data parameters will be sent on the request query string. For POST (and other methods that
+     contain a body), they will be sent as body form parameters, unless the body is explicitly set by {@link
+    #requestBody(String)}, in which case they will be query string parameters.</p>
+
+     @param keyvals a set of key value pairs.
+     @return this Connection, for chaining
      */
     Connection data(String... keyvals);
 
@@ -207,7 +240,7 @@ public interface Connection {
      * @param key the data key
      * @return null if not set
      */
-    KeyVal data(String key);
+    @Nullable KeyVal data(String key);
 
     /**
      * Set a POST (or PUT) request body. Useful when a server expects a plain request body, not a set for URL
@@ -253,6 +286,21 @@ public interface Connection {
      */
     Connection cookies(Map<String, String> cookies);
 
+    /**
+     Provide a custom or pre-filled CookieStore to be used on requests made by this Connection.
+     @param cookieStore a cookie store to use for subsequent requests
+     @return this Connection, for chaining
+     @since 1.14.1
+     */
+    Connection cookieStore(CookieStore cookieStore);
+
+    /**
+     Get the cookie store used by this Connection.
+     @return the cookie store
+     @since 1.14.1
+     */
+    CookieStore cookieStore();
+
     /**
      * Provide an alternate parser to use when parsing the response to a Document. If not set, defaults to the HTML
      * parser, unless the response content-type is XML, in which case the XML parser is used.
@@ -315,8 +363,9 @@ public interface Connection {
     Connection request(Request request);
 
     /**
-     * Get the response, once the request has been executed
+     * Get the response, once the request has been executed.
      * @return response
+     * @throws IllegalArgumentException if called before the response has been executed.
      */
     Response response();
 
@@ -331,11 +380,12 @@ public interface Connection {
      * Common methods for Requests and Responses
      * @param <T> Type of Base, either Request or Response
      */
-    interface Base<T extends Base> {
-
+    @SuppressWarnings("UnusedReturnValue")
+    interface Base<T extends Base<T>> {
         /**
-         * Get the URL
+         * Get the URL of this Request or Response. For redirected responses, this will be the final destination URL.
          * @return URL
+         * @throws IllegalArgumentException if called on a Request that was created without a URL.
          */
         URL url();
 
@@ -347,7 +397,7 @@ public interface Connection {
         T url(URL url);
 
         /**
-         * Get the request method
+         * Get the request method, which defaults to <code>GET</code>
          * @return method
          */
         Method method();
@@ -360,7 +410,8 @@ public interface Connection {
         T method(Method method);
 
         /**
-         * Get the value of a header. This is a simplified header model, where a header may only have one value.
+         * Get the value of a header. If there is more than one header value with the same name, the headers are returned
+         * comma seperated, per <a href="https://www.w3.org/Protocols/rfc2616/rfc2616-sec4.html#sec4.2">rfc2616-sec4</a>.
          * <p>
          * Header names are case insensitive.
          * </p>
@@ -369,16 +420,33 @@ public interface Connection {
          * @see #hasHeader(String)
          * @see #cookie(String)
          */
-        String header(String name);
+        @Nullable String header(String name);
 
         /**
-         * Set a header. This method will overwrite any existing header with the same case insensitive name.
+         * Get the values of a header.
+         * @param name header name, case insensitive.
+         * @return a list of values for this header, or an empty list if not set.
+         */
+        List<String> headers(String name);
+
+        /**
+         * Set a header. This method will overwrite any existing header with the same case insensitive name. (If there
+         * is more than one value for this header, this method will update the first matching header.
          * @param name Name of header
          * @param value Value of header
          * @return this, for chaining
+         * @see #addHeader(String, String)
          */
         T header(String name, String value);
 
+        /**
+         * Add a header. The header will be added regardless of whether a header with the same name already exists.
+         * @param name Name of new header
+         * @param value Value of new header
+         * @return this, for chaining
+         */
+        T addHeader(String name, String value);
+
         /**
          * Check if a header is present
          * @param name name of header (case insensitive)
@@ -395,18 +463,29 @@ public interface Connection {
         boolean hasHeaderWithValue(String name, String value);
 
         /**
-         * Remove a header by name
+         * Remove headers by name. If there is more than one header with this name, they will all be removed.
          * @param name name of header to remove (case insensitive)
          * @return this, for chaining
          */
         T removeHeader(String name);
 
         /**
-         * Retrieve all of the request/response headers as a map
+         * Retrieve all of the request/response header names and corresponding values as a map. For headers with multiple
+         * values, only the first header is returned.
+         * <p>Note that this is a view of the headers only, and changes made to this map will not be reflected in the
+         * request/response object.</p>
          * @return headers
+         * @see #multiHeaders()
+
          */
         Map<String, String> headers();
 
+        /**
+         * Retreive all of the headers, keyed by the header name, and with a list of values per header.
+         * @return a list of multiple values per header.
+         */
+        Map<String, List<String>> multiHeaders();
+
         /**
          * Get a cookie value by name from this request/response.
          * <p>
@@ -416,7 +495,7 @@ public interface Connection {
          * @param name name of cookie to retrieve.
          * @return value of cookie, or null if not set
          */
-        String cookie(String name);
+        @Nullable String cookie(String name);
 
         /**
          * Set a cookie in this request/response.
@@ -450,19 +529,20 @@ public interface Connection {
     /**
      * Represents a HTTP request.
      */
+    @SuppressWarnings("UnusedReturnValue")
     interface Request extends Base<Request> {
         /**
          * Get the proxy used for this request.
          * @return the proxy; <code>null</code> if not enabled.
          */
-        Proxy proxy();
+        @Nullable Proxy proxy();
 
         /**
          * Update the proxy for this request.
          * @param proxy the proxy ot use; <code>null</code> to disable.
          * @return this Request, for chaining
          */
-        Request proxy(Proxy proxy);
+        Request proxy(@Nullable Proxy proxy);
 
         /**
          * Set the HTTP proxy to use for this request.
@@ -540,16 +620,16 @@ public interface Connection {
         Request ignoreContentType(boolean ignoreContentType);
 
         /**
-         * Get the current state of TLS (SSL) certificate validation.
-         * @return true if TLS cert validation enabled
+         * Get the current custom SSL socket factory, if any.
+         * @return custom SSL socket factory if set, null otherwise
          */
-        boolean validateTLSCertificates();
+        @Nullable SSLSocketFactory sslSocketFactory();
 
         /**
-         * Set TLS certificate validation.
-         * @param value set false to ignore TLS (SSL) certificates
+         * Set a custom SSL socket factory.
+         * @param sslSocketFactory SSL socket factory
          */
-        void validateTLSCertificates(boolean value);
+        void sslSocketFactory(SSLSocketFactory sslSocketFactory);
 
         /**
          * Add a data parameter to the request
@@ -572,15 +652,16 @@ public interface Connection {
          * .header("Content-Type", "application/json")
          * .post();</pre></code>
          * If any data key/vals are supplied, they will be sent as URL query params.
+         * @param body to use as the request body. Set to null to clear a previously set body.
          * @return this Request, for chaining
          */
-        Request requestBody(String body);
+        Request requestBody(@Nullable String body);
 
         /**
          * Get the current request body.
          * @return null if not set.
          */
-        String requestBody();
+        @Nullable String requestBody();
 
         /**
          * Specify the parser to use when parsing the document.
@@ -629,9 +710,9 @@ public interface Connection {
 
         /**
          * Get the character set name of the response, derived from the content-type header.
-         * @return character set name
+         * @return character set name if set, <b>null</b> if not
          */
-        String charset();
+        @Nullable String charset();
 
         /**
          * Set / override the response character set. When the document body is parsed it will be with this charset.
@@ -642,12 +723,13 @@ public interface Connection {
 
         /**
          * Get the response content type (e.g. "text/html");
-         * @return the response content type
+         * @return the response content type, or <b>null</b> if one was not set
          */
-        String contentType();
+        @Nullable String contentType();
 
         /**
-         * Parse the body of the response as a Document.
+         * Read and parse the body of the response as a Document. If you intend to parse the same response multiple
+         * times, you should {@link #bufferUp()} first.
          * @return a parsed Document
          * @throws IOException on error
          */
@@ -664,10 +746,27 @@ public interface Connection {
          * @return body bytes
          */
         byte[] bodyAsBytes();
+
+        /**
+         * Read the body of the response into a local buffer, so that {@link #parse()} may be called repeatedly on the
+         * same connection response (otherwise, once the response is read, its InputStream will have been drained and
+         * may not be re-read). Calling {@link #body() } or {@link #bodyAsBytes()} has the same effect.
+         * @return this response, for chaining
+         * @throws UncheckedIOException if an IO exception occurs during buffering.
+         */
+        Response bufferUp();
+
+        /**
+         * Get the body of the response as a (buffered) InputStream. You should close the input stream when you're done with it.
+         * Other body methods (like bufferUp, body, parse, etc) will not work in conjunction with this method.
+         * <p>This method is useful for writing large responses to disk, without buffering them completely into memory first.</p>
+         * @return the response body input stream
+         */
+        BufferedInputStream bodyStream();
     }
 
     /**
-     * A Key Value tuple.
+     * A Key:Value tuple(+), used for form data.
      */
     interface KeyVal {
 
@@ -708,12 +807,27 @@ public interface Connection {
          * Get the input stream associated with this keyval, if any
          * @return input stream if set, or null
          */
-        InputStream inputStream();
+        @Nullable InputStream inputStream();
 
         /**
          * Does this keyval have an input stream?
          * @return true if this keyval does indeed have an input stream
          */
         boolean hasInputStream();
+
+        /**
+         * Set the Content Type header used in the MIME body (aka mimetype) when uploading files.
+         * Only useful if {@link #inputStream(InputStream)} is set.
+         * <p>Will default to {@code application/octet-stream}.</p>
+         * @param contentType the new content type
+         * @return this KeyVal
+         */
+        KeyVal contentType(String contentType);
+
+        /**
+         * Get the current Content Type, or {@code null} if not set.
+         * @return the current Content Type.
+         */
+        @Nullable String contentType();
     }
 }
