@@ -6,17 +6,14 @@ import java.io.IOException;
  A data node, for contents of style, script tags etc, where contents should not show in text().
 
  @author Jonathan Hedley, jonathan@hedley.net */
-public class DataNode extends Node{
-    private static final String DATA_KEY = "data";
+public class DataNode extends LeafNode {
 
     /**
      Create a new DataNode.
      @param data data contents
-     @param baseUri base URI
      */
-    public DataNode(String data, String baseUri) {
-        super(baseUri);
-        attributes.put(DATA_KEY, data);
+    public DataNode(String data) {
+        value = data;
     }
 
     public String nodeName() {
@@ -28,7 +25,7 @@ public class DataNode extends Node{
      @return data
      */
     public String getWholeData() {
-        return attributes.get(DATA_KEY);
+        return coreValue();
     }
 
     /**
@@ -37,7 +34,7 @@ public class DataNode extends Node{
      * @return this node, for chaining
      */
     public DataNode setWholeData(String data) {
-        attributes.put(DATA_KEY, data);
+        coreValue(data);
         return this;
     }
 
@@ -52,14 +49,21 @@ public class DataNode extends Node{
         return outerHtml();
     }
 
+    @Override
+    public DataNode clone() {
+        return (DataNode) super.clone();
+    }
+
     /**
      Create a new DataNode from HTML encoded data.
      @param encodedData encoded data
-     @param baseUri bass URI
+     @param baseUri base URI
      @return new DataNode
+     @deprecated Unused, and will be removed in 1.15.1.
      */
+    @Deprecated
     public static DataNode createFromEncoded(String encodedData, String baseUri) {
         String data = Entities.unescape(encodedData);
-        return new DataNode(data, baseUri);
+        return new DataNode(data);
     }
 }
