@@ -1,11 +1,17 @@
 package org.jsoup.nodes;
 
-import org.jsoup.helper.StringUtil;
+import org.jsoup.Connection;
+import org.jsoup.Jsoup;
+import org.jsoup.helper.DataUtil;
 import org.jsoup.helper.Validate;
+import org.jsoup.internal.StringUtil;
 import org.jsoup.parser.ParseSettings;
+import org.jsoup.parser.Parser;
 import org.jsoup.parser.Tag;
 import org.jsoup.select.Elements;
+import org.jsoup.select.Evaluator;
 
+import javax.annotation.Nullable;
 import java.nio.charset.Charset;
 import java.nio.charset.CharsetEncoder;
 import java.util.ArrayList;
@@ -16,9 +22,11 @@ import java.util.List;
 
  @author Jonathan Hedley, jonathan@hedley.net */
 public class Document extends Element {
+    private @Nullable Connection connection; // the connection this doc was fetched from, if any
     private OutputSettings outputSettings = new OutputSettings();
+    private Parser parser; // the parser used to parse this document
     private QuirksMode quirksMode = QuirksMode.noQuirks;
-    private String location;
+    private final String location;
     private boolean updateMetaCharset = false;
 
     /**
@@ -30,6 +38,7 @@ public class Document extends Element {
     public Document(String baseUri) {
         super(Tag.valueOf("#root", ParseSettings.htmlDefault), baseUri);
         this.location = baseUri;
+        this.parser = Parser.htmlParser(); // default, but overridable
     }
 
     /**
@@ -41,6 +50,7 @@ public class Document extends Element {
         Validate.notNull(baseUri);
 
         Document doc = new Document(baseUri);
+        doc.parser = doc.parser();
         Element html = doc.appendElement("html");
         html.appendElement("head");
         html.appendElement("body");
@@ -51,26 +61,86 @@ public class Document extends Element {
     /**
      * Get the URL this Document was parsed from. If the starting URL is a redirect,
      * this will return the final URL from which the document was served from.
+     * <p>Will return an empty string if the location is unknown (e.g. if parsed from a String).
      * @return location
      */
     public String location() {
-     return location;
-    }
-    
-    /**
-     Accessor to the document's {@code head} element.
-     @return {@code head}
-     */
-    public Element head() {
-        return findFirstElementByTagName("head", this);
+        return location;
     }
 
     /**
-     Accessor to the document's {@code body} element.
-     @return {@code body}
+     Returns the Connection (Request/Response) object that was used to fetch this document, if any; otherwise, a new
+     default Connection object. This can be used to continue a session, preserving settings and cookies, etc.
+     @return the Connection (session) associated with this Document, or an empty one otherwise.
+     @see Connection#newRequest()
+     */
+    public Connection connection() {
+        if (connection == null)
+            return Jsoup.newSession();
+        else
+            return connection;
+    }
+
+    /**
+     * Returns this Document's doctype.
+     * @return document type, or null if not set
+     */
+    public @Nullable DocumentType documentType() {
+        for (Node node : childNodes) {
+            if (node instanceof DocumentType)
+                return (DocumentType) node;
+            else if (!(node instanceof LeafNode)) // scans forward across comments, text, processing instructions etc
+                break;
+        }
+        return null;
+        // todo - add a set document type?
+    }
+
+    /**
+     Find the root HTML element, or create it if it doesn't exist.
+     @return the root HTML element.
+     */
+    private Element htmlEl() {
+        for (Element el: childElementsList()) {
+            if (el.normalName().equals("html"))
+                return el;
+        }
+        return appendElement("html");
+    }
+
+    /**
+     Get this document's {@code head} element.
+     <p>
+     As a side-effect, if this Document does not already have a HTML structure, it will be created. If you do not want
+     that, use {@code #selectFirst("head")} instead.
+
+     @return {@code head} element.
+     */
+    public Element head() {
+        Element html = htmlEl();
+        for (Element el: html.childElementsList()) {
+            if (el.normalName().equals("head"))
+                return el;
+        }
+        return html.prependElement("head");
+    }
+
+    /**
+     Get this document's {@code <body>} or {@code <frameset>} element.
+     <p>
+     As a <b>side-effect</b>, if this Document does not already have a HTML structure, it will be created with a {@code
+    <body>} element. If you do not want that, use {@code #selectFirst("body")} instead.
+
+     @return {@code body} element for documents with a {@code <body>}, a new {@code <body>} element if the document
+     had no contents, or the outermost {@code <frameset> element} for frameset documents.
      */
     public Element body() {
-        return findFirstElementByTagName("body", this);
+        Element html = htmlEl();
+        for (Element el: html.childElementsList()) {
+            if ("body".equals(el.normalName()) || "frameset".equals(el.normalName()))
+                return el;
+        }
+        return html.appendElement("body");
     }
 
     /**
@@ -79,9 +149,10 @@ public class Document extends Element {
      */
     public String title() {
         // title is a preserve whitespace tag (for document output), but normalised here
-        Element titleEl = getElementsByTag("title").first();
+        Element titleEl = head().selectFirst(titleEval);
         return titleEl != null ? StringUtil.normaliseWhitespace(titleEl.text()).trim() : "";
     }
+    private static final Evaluator titleEval = new Evaluator.Tag("title");
 
     /**
      Set the document's {@code title} element. Updates the existing element, or adds {@code title} to {@code head} if
@@ -90,12 +161,10 @@ public class Document extends Element {
      */
     public void title(String title) {
         Validate.notNull(title);
-        Element titleEl = getElementsByTag("title").first();
-        if (titleEl == null) { // add to head
-            head().appendElement("title").text(title);
-        } else {
-            titleEl.text(title);
-        }
+        Element titleEl = head().selectFirst(titleEval);
+        if (titleEl == null) // add to head
+            titleEl = head().appendElement("title");
+        titleEl.text(title);
     }
 
     /**
@@ -113,17 +182,13 @@ public class Document extends Element {
      @return this document after normalisation
      */
     public Document normalise() {
-        Element htmlEl = findFirstElementByTagName("html", this);
-        if (htmlEl == null)
-            htmlEl = appendElement("html");
-        if (head() == null)
-            htmlEl.prependElement("head");
-        if (body() == null)
-            htmlEl.appendElement("body");
+        Element htmlEl = htmlEl(); // these all create if not found
+        Element head = head();
+        body();
 
         // pull text nodes out of root, html, and head els, and push into body. non-text nodes are already taken care
         // of. do in inverse order to maintain text order.
-        normaliseTextNodes(head());
+        normaliseTextNodes(head);
         normaliseTextNodes(htmlEl);
         normaliseTextNodes(this);
 
@@ -137,7 +202,7 @@ public class Document extends Element {
 
     // does not recurse.
     private void normaliseTextNodes(Element element) {
-        List<Node> toMove = new ArrayList<Node>();
+        List<Node> toMove = new ArrayList<>();
         for (Node node: element.childNodes) {
             if (node instanceof TextNode) {
                 TextNode tn = (TextNode) node;
@@ -149,7 +214,7 @@ public class Document extends Element {
         for (int i = toMove.size()-1; i >= 0; i--) {
             Node node = toMove.get(i);
             element.removeChild(node);
-            body().prependChild(new TextNode(" ", ""));
+            body().prependChild(new TextNode(" "));
             body().prependChild(node);
         }
     }
@@ -159,11 +224,10 @@ public class Document extends Element {
         Elements elements = this.getElementsByTag(tag);
         Element master = elements.first(); // will always be available as created above if not existent
         if (elements.size() > 1) { // dupes, move contents to master
-            List<Node> toMove = new ArrayList<Node>();
+            List<Node> toMove = new ArrayList<>();
             for (int i = 1; i < elements.size(); i++) {
                 Node dupe = elements.get(i);
-                for (Node node : dupe.childNodes)
-                    toMove.add(node);
+                toMove.addAll(dupe.ensureChildNodes());
                 dupe.remove();
             }
 
@@ -171,25 +235,11 @@ public class Document extends Element {
                 master.appendChild(dupe);
         }
         // ensure parented by <html>
-        if (!master.parent().equals(htmlEl)) {
+        if (master.parent() != null && !master.parent().equals(htmlEl)) {
             htmlEl.appendChild(master); // includes remove()            
         }
     }
 
-    // fast method to get first by tag name, used for html, head, body finders
-    private Element findFirstElementByTagName(String tag, Node node) {
-        if (node.nodeName().equals(tag))
-            return (Element) node;
-        else {
-            for (Node child: node.childNodes) {
-                Element found = findFirstElementByTagName(tag, child);
-                if (found != null)
-                    return found;
-            }
-        }
-        return null;
-    }
-
     @Override
     public String outerHtml() {
         return super.html(); // no outer wrapper tag
@@ -313,46 +363,31 @@ public class Document extends Element {
             OutputSettings.Syntax syntax = outputSettings().syntax();
 
             if (syntax == OutputSettings.Syntax.html) {
-                Element metaCharset = select("meta[charset]").first();
-
+                Element metaCharset = selectFirst("meta[charset]");
                 if (metaCharset != null) {
                     metaCharset.attr("charset", charset().displayName());
                 } else {
-                    Element head = head();
-
-                    if (head != null) {
-                        head.appendElement("meta").attr("charset", charset().displayName());
-                    }
+                    head().appendElement("meta").attr("charset", charset().displayName());
                 }
-
-                // Remove obsolete elements
-                select("meta[name=charset]").remove();
+                select("meta[name=charset]").remove(); // Remove obsolete elements
             } else if (syntax == OutputSettings.Syntax.xml) {
-                Node node = childNodes().get(0);
-
+                Node node = ensureChildNodes().get(0);
                 if (node instanceof XmlDeclaration) {
                     XmlDeclaration decl = (XmlDeclaration) node;
-
                     if (decl.name().equals("xml")) {
                         decl.attr("encoding", charset().displayName());
-
-                        final String version = decl.attr("version");
-
-                        if (version != null) {
+                        if (decl.hasAttr("version"))
                             decl.attr("version", "1.0");
-                        }
                     } else {
-                        decl = new XmlDeclaration("xml", baseUri, false);
+                        decl = new XmlDeclaration("xml", false);
                         decl.attr("version", "1.0");
                         decl.attr("encoding", charset().displayName());
-
                         prependChild(decl);
                     }
                 } else {
-                    XmlDeclaration decl = new XmlDeclaration("xml", baseUri, false);
+                    XmlDeclaration decl = new XmlDeclaration("xml", false);
                     decl.attr("version", "1.0");
                     decl.attr("encoding", charset().displayName());
-
                     prependChild(decl);
                 }
             }
@@ -370,7 +405,10 @@ public class Document extends Element {
         public enum Syntax {html, xml}
 
         private Entities.EscapeMode escapeMode = Entities.EscapeMode.base;
-        private Charset charset = Charset.forName("UTF-8");
+        private Charset charset = DataUtil.UTF_8;
+        private final ThreadLocal<CharsetEncoder> encoderThreadLocal = new ThreadLocal<>(); // initialized by start of OuterHtmlVisitor
+        @Nullable Entities.CoreCharset coreCharset; // fast encoders for ascii and utf8
+
         private boolean prettyPrint = true;
         private boolean outline = false;
         private int indentAmount = 1;
@@ -433,8 +471,17 @@ public class Document extends Element {
             return this;
         }
 
+        CharsetEncoder prepareEncoder() {
+            // created at start of OuterHtmlVisitor so each pass has own encoder, so OutputSettings can be shared among threads
+            CharsetEncoder encoder = charset.newEncoder();
+            encoderThreadLocal.set(encoder);
+            coreCharset = Entities.CoreCharset.byName(encoder.charset().name());
+            return encoder;
+        }
+
         CharsetEncoder encoder() {
-            return charset.newEncoder();
+            CharsetEncoder encoder = encoderThreadLocal.get();
+            return encoder != null ? encoder : prepareEncoder();
         }
 
         /**
@@ -559,4 +606,38 @@ public class Document extends Element {
         this.quirksMode = quirksMode;
         return this;
     }
+
+    /**
+     * Get the parser that was used to parse this document.
+     * @return the parser
+     */
+    public Parser parser() {
+        return parser;
+    }
+
+    /**
+     * Set the parser used to create this document. This parser is then used when further parsing within this document
+     * is required.
+     * @param parser the configured parser to use when further parsing is required for this document.
+     * @return this document, for chaining.
+     */
+    public Document parser(Parser parser) {
+        this.parser = parser;
+        return this;
+    }
+
+    /**
+     Set the Connection used to fetch this document. This Connection is used as a session object when further requests are
+     made (e.g. when a form is submitted).
+
+     @param connection to set
+     @return this document, for chaining
+     @see Connection#newRequest()
+     @since 1.14.1
+     */
+    public Document connection(Connection connection) {
+        Validate.notNull(connection);
+        this.connection = connection;
+        return this;
+    }
 }
