@@ -9,12 +9,22 @@ import java.util.ArrayList;
  */
 public class ParseErrorList extends ArrayList<ParseError>{
     private static final int INITIAL_CAPACITY = 16;
+    private final int initialCapacity;
     private final int maxSize;
     
     ParseErrorList(int initialCapacity, int maxSize) {
         super(initialCapacity);
+        this.initialCapacity = initialCapacity;
         this.maxSize = maxSize;
     }
+
+    /**
+     Create a new ParseErrorList with the same settings, but no errors in the list
+     @param copy initial and max size details to copy
+     */
+    ParseErrorList(ParseErrorList copy) {
+        this(copy.initialCapacity, copy.maxSize);
+    }
     
     boolean canAddError() {
         return size() < maxSize;
