@@ -1,9 +1,11 @@
 package org.jsoup.parser;
 
 import org.jsoup.helper.Validate;
-import org.jsoup.nodes.Attribute;
 import org.jsoup.nodes.Attributes;
-import org.jsoup.nodes.BooleanAttribute;
+
+import javax.annotation.Nullable;
+
+import static org.jsoup.internal.Normalizer.lowerCase;
 
 /**
  * Parse tokens for the Tokeniser.
@@ -70,75 +72,113 @@ abstract class Token {
         public boolean isForceQuirks() {
             return forceQuirks;
         }
+
+        @Override
+        public String toString() {
+            return "<!doctype " + getName() + ">";
+        }
     }
 
     static abstract class Tag extends Token {
-        protected String tagName;
-        protected String normalName; // lc version of tag name, for case insensitive tree build
-        private String pendingAttributeName; // attribute names are generally caught in one hop, not accumulated
-        private StringBuilder pendingAttributeValue = new StringBuilder(); // but values are accumulated, from e.g. & in hrefs
-        private String pendingAttributeValueS; // try to get attr vals in one shot, vs Builder
-        private boolean hasEmptyAttributeValue = false; // distinguish boolean attribute from empty string value
-        private boolean hasPendingAttributeValue = false;
+        @Nullable protected String tagName;
+        @Nullable protected String normalName; // lc version of tag name, for case insensitive tree build
+
+        private final StringBuilder attrName = new StringBuilder(); // try to get attr names and vals in one shot, vs Builder
+        @Nullable private String attrNameS;
+        private boolean hasAttrName = false;
+
+        private final StringBuilder attrValue = new StringBuilder();
+        @Nullable private String attrValueS;
+        private boolean hasAttrValue = false;
+        private boolean hasEmptyAttrValue = false; // distinguish boolean attribute from empty string value
+
         boolean selfClosing = false;
-        Attributes attributes; // start tags get attributes on construction. End tags get attributes on first new attribute (but only for parser convenience, not used).
+        @Nullable Attributes attributes; // start tags get attributes on construction. End tags get attributes on first new attribute (but only for parser convenience, not used).
 
         @Override
         Tag reset() {
             tagName = null;
             normalName = null;
-            pendingAttributeName = null;
-            reset(pendingAttributeValue);
-            pendingAttributeValueS = null;
-            hasEmptyAttributeValue = false;
-            hasPendingAttributeValue = false;
+            reset(attrName);
+            attrNameS = null;
+            hasAttrName = false;
+            reset(attrValue);
+            attrValueS = null;
+            hasEmptyAttrValue = false;
+            hasAttrValue = false;
             selfClosing = false;
             attributes = null;
             return this;
         }
 
+        /* Limits runaway crafted HTML from spewing attributes and getting a little sluggish in ensureCapacity.
+        Real-world HTML will P99 around 8 attributes, so plenty of headroom. Implemented here and not in the Attributes
+        object so that API users can add more if ever required. */
+        private static final int MaxAttributes = 512;
+
         final void newAttribute() {
             if (attributes == null)
                 attributes = new Attributes();
 
-            if (pendingAttributeName != null) {
-                Attribute attribute;
-                if (hasPendingAttributeValue)
-                    attribute = new Attribute(pendingAttributeName,
-                        pendingAttributeValue.length() > 0 ? pendingAttributeValue.toString() : pendingAttributeValueS);
-                else if (hasEmptyAttributeValue)
-                    attribute = new Attribute(pendingAttributeName, "");
-                else
-                    attribute = new BooleanAttribute(pendingAttributeName);
-                attributes.put(attribute);
+            if (hasAttrName && attributes.size() < MaxAttributes) {
+                // the tokeniser has skipped whitespace control chars, but trimming could collapse to empty for other control codes, so verify here
+                String name = attrName.length() > 0 ? attrName.toString() : attrNameS;
+                name = name.trim();
+                if (name.length() > 0) {
+                    String value;
+                    if (hasAttrValue)
+                        value = attrValue.length() > 0 ? attrValue.toString() : attrValueS;
+                    else if (hasEmptyAttrValue)
+                        value = "";
+                    else
+                        value = null;
+                    // note that we add, not put. So that the first is kept, and rest are deduped, once in a context where case sensitivity is known (the appropriate tree builder).
+                    attributes.add(name, value);
+                }
             }
-            pendingAttributeName = null;
-            hasEmptyAttributeValue = false;
-            hasPendingAttributeValue = false;
-            reset(pendingAttributeValue);
-            pendingAttributeValueS = null;
+            reset(attrName);
+            attrNameS = null;
+            hasAttrName = false;
+
+            reset(attrValue);
+            attrValueS = null;
+            hasAttrValue = false;
+            hasEmptyAttrValue = false;
+        }
+
+        final boolean hasAttributes() {
+            return attributes != null;
+        }
+
+        final boolean hasAttribute(String key) {
+            return attributes != null && attributes.hasKey(key);
         }
 
         final void finaliseTag() {
             // finalises for emit
-            if (pendingAttributeName != null) {
-                // todo: check if attribute name exists; if so, drop and error
+            if (hasAttrName) {
                 newAttribute();
             }
         }
 
+        /** Preserves case */
         final String name() { // preserves case, for input into Tag.valueOf (which may drop case)
             Validate.isFalse(tagName == null || tagName.length() == 0);
             return tagName;
         }
 
-        final String normalName() { // loses case, used in tree building for working out where in tree it should go
+        /** Lower case */
+        final String normalName() { // lower case, used in tree building for working out where in tree it should go
             return normalName;
         }
 
+        final String toStringName() {
+            return tagName != null ? tagName : "[unset]";
+        }
+
         final Tag name(String name) {
             tagName = name;
-            normalName = name.toLowerCase();
+            normalName = ParseSettings.normalName(tagName);
             return this;
         }
 
@@ -146,15 +186,12 @@ abstract class Token {
             return selfClosing;
         }
 
-        @SuppressWarnings({"TypeMayBeWeakened"})
-        final Attributes getAttributes() {
-            return attributes;
-        }
-
         // these appenders are rarely hit in not null state-- caused by null chars.
         final void appendTagName(String append) {
+            // might have null chars - need to replace with null replacement character
+            append = append.replace(TokeniserState.nullChar, Tokeniser.replacementChar);
             tagName = tagName == null ? append : tagName.concat(append);
-            normalName = tagName.toLowerCase();
+            normalName = ParseSettings.normalName(tagName);
         }
 
         final void appendTagName(char append) {
@@ -162,81 +199,100 @@ abstract class Token {
         }
 
         final void appendAttributeName(String append) {
-            pendingAttributeName = pendingAttributeName == null ? append : pendingAttributeName.concat(append);
+            // might have null chars because we eat in one pass - need to replace with null replacement character
+            append = append.replace(TokeniserState.nullChar, Tokeniser.replacementChar);
+
+            ensureAttrName();
+            if (attrName.length() == 0) {
+                attrNameS = append;
+            } else {
+                attrName.append(append);
+            }
         }
 
         final void appendAttributeName(char append) {
-            appendAttributeName(String.valueOf(append));
+            ensureAttrName();
+            attrName.append(append);
         }
 
         final void appendAttributeValue(String append) {
-            ensureAttributeValue();
-            if (pendingAttributeValue.length() == 0) {
-                pendingAttributeValueS = append;
+            ensureAttrValue();
+            if (attrValue.length() == 0) {
+                attrValueS = append;
             } else {
-                pendingAttributeValue.append(append);
+                attrValue.append(append);
             }
         }
 
         final void appendAttributeValue(char append) {
-            ensureAttributeValue();
-            pendingAttributeValue.append(append);
+            ensureAttrValue();
+            attrValue.append(append);
         }
 
         final void appendAttributeValue(char[] append) {
-            ensureAttributeValue();
-            pendingAttributeValue.append(append);
+            ensureAttrValue();
+            attrValue.append(append);
         }
 
         final void appendAttributeValue(int[] appendCodepoints) {
-            ensureAttributeValue();
+            ensureAttrValue();
             for (int codepoint : appendCodepoints) {
-                pendingAttributeValue.appendCodePoint(codepoint);
+                attrValue.appendCodePoint(codepoint);
             }
         }
         
         final void setEmptyAttributeValue() {
-            hasEmptyAttributeValue = true;
+            hasEmptyAttrValue = true;
         }
 
-        private void ensureAttributeValue() {
-            hasPendingAttributeValue = true;
+        private void ensureAttrName() {
+            hasAttrName = true;
             // if on second hit, we'll need to move to the builder
-            if (pendingAttributeValueS != null) {
-                pendingAttributeValue.append(pendingAttributeValueS);
-                pendingAttributeValueS = null;
+            if (attrNameS != null) {
+                attrName.append(attrNameS);
+                attrNameS = null;
             }
         }
+
+        private void ensureAttrValue() {
+            hasAttrValue = true;
+            // if on second hit, we'll need to move to the builder
+            if (attrValueS != null) {
+                attrValue.append(attrValueS);
+                attrValueS = null;
+            }
+        }
+
+        @Override
+        abstract public String toString();
     }
 
     final static class StartTag extends Tag {
         StartTag() {
             super();
-            attributes = new Attributes();
             type = TokenType.StartTag;
         }
 
         @Override
         Tag reset() {
             super.reset();
-            attributes = new Attributes();
-            // todo - would prefer these to be null, but need to check Element assertions
+            attributes = null;
             return this;
         }
 
         StartTag nameAttr(String name, Attributes attributes) {
             this.tagName = name;
             this.attributes = attributes;
-            normalName = tagName.toLowerCase();
+            normalName = ParseSettings.normalName(tagName);
             return this;
         }
 
         @Override
         public String toString() {
-            if (attributes != null && attributes.size() > 0)
-                return "<" + name() + " " + attributes.toString() + ">";
+            if (hasAttributes() && attributes.size() > 0)
+                return "<" + toStringName() + " " + attributes.toString() + ">";
             else
-                return "<" + name() + ">";
+                return "<" + toStringName() + ">";
         }
     }
 
@@ -248,17 +304,19 @@ abstract class Token {
 
         @Override
         public String toString() {
-            return "</" + name() + ">";
+            return "</" + toStringName() + ">";
         }
     }
 
     final static class Comment extends Token {
-        final StringBuilder data = new StringBuilder();
+        private final StringBuilder data = new StringBuilder();
+        private String dataS; // try to get in one shot
         boolean bogus = false;
 
         @Override
         Token reset() {
             reset(data);
+            dataS = null;
             bogus = false;
             return this;
         }
@@ -268,7 +326,31 @@ abstract class Token {
         }
 
         String getData() {
-            return data.toString();
+            return dataS != null ? dataS : data.toString();
+        }
+
+        final Comment append(String append) {
+            ensureData();
+            if (data.length() == 0) {
+                dataS = append;
+            } else {
+                data.append(append);
+            }
+            return this;
+        }
+
+        final Comment append(char append) {
+            ensureData();
+            data.append(append);
+            return this;
+        }
+
+        private void ensureData() {
+            // if on second hit, we'll need to move to the builder
+            if (dataS != null) {
+                data.append(dataS);
+                dataS = null;
+            }
         }
 
         @Override
@@ -277,7 +359,7 @@ abstract class Token {
         }
     }
 
-    final static class Character extends Token {
+    static class Character extends Token {
         private String data;
 
         Character() {
@@ -306,6 +388,19 @@ abstract class Token {
         }
     }
 
+    final static class CData extends Character {
+        CData(String data) {
+            super();
+            this.data(data);
+        }
+
+        @Override
+        public String toString() {
+            return "<![CDATA[" + getData() + "]]>";
+        }
+
+    }
+
     final static class EOF extends Token {
         EOF() {
             type = Token.TokenType.EOF;
@@ -315,6 +410,11 @@ abstract class Token {
         Token reset() {
             return this;
         }
+
+        @Override
+        public String toString() {
+            return "";
+        }
     }
 
     final boolean isDoctype() {
@@ -353,6 +453,10 @@ abstract class Token {
         return type == TokenType.Character;
     }
 
+    final boolean isCData() {
+        return this instanceof CData;
+    }
+
     final Character asCharacter() {
         return (Character) this;
     }
@@ -361,12 +465,12 @@ abstract class Token {
         return type == TokenType.EOF;
     }
 
-    enum TokenType {
+    public enum TokenType {
         Doctype,
         StartTag,
         EndTag,
         Comment,
-        Character,
+        Character, // note no CData - treated in builder as an extension of Character
         EOF
     }
 }
