@@ -1,8 +1,10 @@
 package org.jsoup.parser;
 
 import org.jsoup.helper.Validate;
+import org.jsoup.internal.StringUtil;
 import org.jsoup.nodes.Entities;
 
+import javax.annotation.Nullable;
 import java.util.Arrays;
 
 /**
@@ -12,12 +14,24 @@ final class Tokeniser {
     static final char replacementChar = '\uFFFD'; // replaces null character
     private static final char[] notCharRefCharsSorted = new char[]{'\t', '\n', '\r', '\f', ' ', '<', '&'};
 
+    // Some illegal character escapes are parsed by browsers as windows-1252 instead. See issue #1034
+    // https://html.spec.whatwg.org/multipage/parsing.html#numeric-character-reference-end-state
+    static final int win1252ExtensionsStart = 0x80;
+    static final int[] win1252Extensions = new int[] {
+            // we could build this manually, but Windows-1252 is not a standard java charset so that could break on
+            // some platforms - this table is verified with a test
+            0x20AC, 0x0081, 0x201A, 0x0192, 0x201E, 0x2026, 0x2020, 0x2021,
+            0x02C6, 0x2030, 0x0160, 0x2039, 0x0152, 0x008D, 0x017D, 0x008F,
+            0x0090, 0x2018, 0x2019, 0x201C, 0x201D, 0x2022, 0x2013, 0x2014,
+            0x02DC, 0x2122, 0x0161, 0x203A, 0x0153, 0x009D, 0x017E, 0x0178,
+    };
+
     static {
         Arrays.sort(notCharRefCharsSorted);
     }
 
-    private CharacterReader reader; // html input
-    private ParseErrorList errors; // errors found while tokenising
+    private final CharacterReader reader; // html input
+    private final ParseErrorList errors; // errors found while tokenising
 
     private TokeniserState state = TokeniserState.Data; // current tokenisation state
     private Token emitPending; // the token we are about to emit on next read
@@ -33,7 +47,7 @@ final class Tokeniser {
     Token.Doctype doctypePending = new Token.Doctype(); // doctype building up
     Token.Comment commentPending = new Token.Comment(); // comment building up
     private String lastStartTag; // the last start tag emitted, to test appropriate end tag
-    private boolean selfClosingFlagAcknowledged = true;
+    @Nullable private String lastStartCloseSeq; // "</" + lastStartTag, so we can quickly check for that in RCData
 
     Tokeniser(CharacterReader reader, ParseErrorList errors) {
         this.reader = reader;
@@ -41,18 +55,15 @@ final class Tokeniser {
     }
 
     Token read() {
-        if (!selfClosingFlagAcknowledged) {
-            error("Self closing flag not acknowledged");
-            selfClosingFlagAcknowledged = true;
+        while (!isEmitPending) {
+            state.read(this, reader);
         }
 
-        while (!isEmitPending)
-            state.read(this, reader);
-
         // if emit is pending, a non-character token was found: return any chars in buffer, and leave token for next read:
-        if (charsBuilder.length() > 0) {
-            String str = charsBuilder.toString();
-            charsBuilder.delete(0, charsBuilder.length());
+        final StringBuilder cb = this.charsBuilder;
+        if (cb.length() != 0) {
+            String str = cb.toString();
+            cb.delete(0, cb.length());
             charsString = null;
             return charPending.data(str);
         } else if (charsString != null) {
@@ -66,7 +77,7 @@ final class Tokeniser {
     }
 
     void emit(Token token) {
-        Validate.isFalse(isEmitPending, "There is an unread token pending!");
+        Validate.isFalse(isEmitPending);
 
         emitPending = token;
         isEmitPending = true;
@@ -74,12 +85,11 @@ final class Tokeniser {
         if (token.type == Token.TokenType.StartTag) {
             Token.StartTag startTag = (Token.StartTag) token;
             lastStartTag = startTag.tagName;
-            if (startTag.selfClosing)
-                selfClosingFlagAcknowledged = false;
+            lastStartCloseSeq = null; // only lazy inits
         } else if (token.type == Token.TokenType.EndTag) {
             Token.EndTag endTag = (Token.EndTag) token;
-            if (endTag.attributes != null)
-                error("Attributes incorrectly present on end tag");
+            if (endTag.hasAttributes())
+                error("Attributes incorrectly present on end tag [/%s]", endTag.normalName());
         }
     }
 
@@ -97,12 +107,37 @@ final class Tokeniser {
         }
     }
 
+    // variations to limit need to create temp strings
+    void emit(final StringBuilder str) {
+        if (charsString == null) {
+            charsString = str.toString();
+        }
+        else {
+            if (charsBuilder.length() == 0) {
+                charsBuilder.append(charsString);
+            }
+            charsBuilder.append(str);
+        }
+    }
+
+    void emit(char c) {
+        if (charsString == null) {
+            charsString = String.valueOf(c);
+        }
+        else {
+            if (charsBuilder.length() == 0) {
+                charsBuilder.append(charsString);
+            }
+            charsBuilder.append(c);
+        }
+    }
+
     void emit(char[] chars) {
         emit(String.valueOf(chars));
     }
 
-    void emit(char c) {
-        emit(String.valueOf(c));
+    void emit(int[] codepoints) {
+        emit(new String(codepoints, 0, codepoints.length));
     }
 
     TokeniserState getState() {
@@ -118,12 +153,9 @@ final class Tokeniser {
         this.state = state;
     }
 
-    void acknowledgeSelfClosingFlag() {
-        selfClosingFlagAcknowledged = true;
-    }
-
-    final private char[] charRefHolder = new char[1]; // holder to not have to keep creating arrays
-    char[] consumeCharacterReference(Character additionalAllowedCharacter, boolean inAttribute) {
+    final private int[] codepointHolder = new int[1]; // holder to not have to keep creating arrays
+    final private int[] multipointHolder = new int[2];
+    @Nullable int[] consumeCharacterReference(Character additionalAllowedCharacter, boolean inAttribute) {
         if (reader.isEmpty())
             return null;
         if (additionalAllowedCharacter != null && additionalAllowedCharacter == reader.current())
@@ -131,7 +163,7 @@ final class Tokeniser {
         if (reader.matchesAnySorted(notCharRefCharsSorted))
             return null;
 
-        final char[] charRef = charRefHolder;
+        final int[] codeRef = codepointHolder;
         reader.mark();
         if (reader.matchConsume("#")) { // numbered
             boolean isHexMode = reader.matchConsumeIgnoreCase("X");
@@ -141,27 +173,31 @@ final class Tokeniser {
                 reader.rewindToMark();
                 return null;
             }
+
+            reader.unmark();
             if (!reader.matchConsume(";"))
-                characterReferenceError("missing semicolon"); // missing semi
+                characterReferenceError("missing semicolon on [&#%s]", numRef); // missing semi
             int charval = -1;
             try {
                 int base = isHexMode ? 16 : 10;
                 charval = Integer.valueOf(numRef, base);
-            } catch (NumberFormatException e) {
+            } catch (NumberFormatException ignored) {
             } // skip
             if (charval == -1 || (charval >= 0xD800 && charval <= 0xDFFF) || charval > 0x10FFFF) {
-                characterReferenceError("character outside of valid range");
-                charRef[0] = replacementChar;
-                return charRef;
+                characterReferenceError("character [%s] outside of valid range", charval);
+                codeRef[0] = replacementChar;
             } else {
+                // fix illegal unicode characters to match browser behavior
+                if (charval >= win1252ExtensionsStart && charval < win1252ExtensionsStart + win1252Extensions.length) {
+                    characterReferenceError("character [%s] is not a valid unicode code point", charval);
+                    charval = win1252Extensions[charval - win1252ExtensionsStart];
+                }
+
                 // todo: implement number replacement table
                 // todo: check for extra illegal unicode points as parse errors
-                if (charval < Character.MIN_SUPPLEMENTARY_CODE_POINT) {
-                    charRef[0] = (char) charval;
-                    return charRef;
-                } else
-                return Character.toChars(charval);
+                codeRef[0] = charval;
             }
+            return codeRef;
         } else { // named
             // get as many letters as possible, and look for matching entities.
             String nameRef = reader.consumeLetterThenDigitSequence();
@@ -172,7 +208,7 @@ final class Tokeniser {
             if (!found) {
                 reader.rewindToMark();
                 if (looksLegit) // named with semicolon
-                    characterReferenceError(String.format("invalid named referenece '%s'", nameRef));
+                    characterReferenceError("invalid named reference [%s]", nameRef);
                 return null;
             }
             if (inAttribute && (reader.matchesLetter() || reader.matchesDigit() || reader.matchesAny('=', '-', '_'))) {
@@ -180,10 +216,20 @@ final class Tokeniser {
                 reader.rewindToMark();
                 return null;
             }
+
+            reader.unmark();
             if (!reader.matchConsume(";"))
-                characterReferenceError("missing semicolon"); // missing semi
-            charRef[0] = Entities.getCharacterByName(nameRef);
-            return charRef;
+                characterReferenceError("missing semicolon on [&%s]", nameRef); // missing semi
+            int numChars = Entities.codepointsForName(nameRef, multipointHolder);
+            if (numChars == 1) {
+                codeRef[0] = multipointHolder[0];
+                return codeRef;
+            } else if (numChars ==2) {
+                return multipointHolder;
+            } else {
+                Validate.fail("Unexpected characters returned for " + nameRef);
+                return multipointHolder;
+            }
         }
     }
 
@@ -205,6 +251,11 @@ final class Tokeniser {
         emit(commentPending);
     }
 
+    void createBogusCommentPending() {
+        commentPending.reset();
+        commentPending.bogus = true;
+    }
+
     void createDoctypePending() {
         doctypePending.reset();
     }
@@ -218,33 +269,43 @@ final class Tokeniser {
     }
 
     boolean isAppropriateEndTagToken() {
-        return lastStartTag != null && tagPending.tagName.equals(lastStartTag);
+        return lastStartTag != null && tagPending.name().equalsIgnoreCase(lastStartTag);
     }
 
     String appropriateEndTagName() {
-        if (lastStartTag == null)
-            return null;
-        return lastStartTag;
+        return lastStartTag; // could be null
+    }
+
+    /** Returns the closer sequence {@code </lastStart} */
+    String appropriateEndTagSeq() {
+        if (lastStartCloseSeq == null) // reset on start tag emit
+            lastStartCloseSeq = "</" + lastStartTag;
+        return lastStartCloseSeq;
     }
 
     void error(TokeniserState state) {
         if (errors.canAddError())
-            errors.add(new ParseError(reader.pos(), "Unexpected character '%s' in input state [%s]", reader.current(), state));
+            errors.add(new ParseError(reader, "Unexpected character '%s' in input state [%s]", reader.current(), state));
     }
 
     void eofError(TokeniserState state) {
         if (errors.canAddError())
-            errors.add(new ParseError(reader.pos(), "Unexpectedly reached end of file (EOF) in input state [%s]", state));
+            errors.add(new ParseError(reader, "Unexpectedly reached end of file (EOF) in input state [%s]", state));
     }
 
-    private void characterReferenceError(String message) {
+    private void characterReferenceError(String message, Object... args) {
         if (errors.canAddError())
-            errors.add(new ParseError(reader.pos(), "Invalid character reference: %s", message));
+            errors.add(new ParseError(reader, String.format("Invalid character reference: " + message, args)));
     }
 
-    private void error(String errorMsg) {
+    void error(String errorMsg) {
         if (errors.canAddError())
-            errors.add(new ParseError(reader.pos(), errorMsg));
+            errors.add(new ParseError(reader, errorMsg));
+    }
+
+    void error(String errorMsg, Object... args) {
+        if (errors.canAddError())
+            errors.add(new ParseError(reader, errorMsg, args));
     }
 
     boolean currentNodeInHtmlNS() {
@@ -256,22 +317,26 @@ final class Tokeniser {
 
     /**
      * Utility method to consume reader and unescape entities found within.
-     * @param inAttribute
+     * @param inAttribute if the text to be unescaped is in an attribute
      * @return unescaped string from reader
      */
     String unescapeEntities(boolean inAttribute) {
-        StringBuilder builder = new StringBuilder();
+        StringBuilder builder = StringUtil.borrowBuilder();
         while (!reader.isEmpty()) {
             builder.append(reader.consumeTo('&'));
             if (reader.matches('&')) {
                 reader.consume();
-                char[] c = consumeCharacterReference(null, inAttribute);
+                int[] c = consumeCharacterReference(null, inAttribute);
                 if (c == null || c.length==0)
                     builder.append('&');
-                else
-                    builder.append(c);
+                else {
+                    builder.appendCodePoint(c[0]);
+                    if (c.length == 2)
+                        builder.appendCodePoint(c[1]);
+                }
+
             }
         }
-        return builder.toString();
+        return StringUtil.releaseBuilder(builder);
     }
 }
