@@ -1,17 +1,24 @@
 package org.jsoup.select;
 
 import org.jsoup.helper.Validate;
+import org.jsoup.internal.StringUtil;
 import org.jsoup.nodes.Comment;
 import org.jsoup.nodes.Document;
 import org.jsoup.nodes.DocumentType;
 import org.jsoup.nodes.Element;
 import org.jsoup.nodes.Node;
+import org.jsoup.nodes.PseudoTextElement;
+import org.jsoup.nodes.TextNode;
 import org.jsoup.nodes.XmlDeclaration;
 
 import java.util.List;
 import java.util.regex.Matcher;
 import java.util.regex.Pattern;
 
+import static org.jsoup.internal.Normalizer.lowerCase;
+import static org.jsoup.internal.Normalizer.normalize;
+import static org.jsoup.internal.StringUtil.normaliseWhitespace;
+
 
 /**
  * Evaluates that an element matches the selector.
@@ -34,7 +41,7 @@ public abstract class Evaluator {
      * Evaluator for tag name
      */
     public static final class Tag extends Evaluator {
-        private String tagName;
+        private final String tagName;
 
         public Tag(String tagName) {
             this.tagName = tagName;
@@ -42,7 +49,7 @@ public abstract class Evaluator {
 
         @Override
         public boolean matches(Element root, Element element) {
-            return (element.tagName().equalsIgnoreCase(tagName));
+            return (element.normalName().equals(tagName));
         }
 
         @Override
@@ -56,7 +63,7 @@ public abstract class Evaluator {
      * Evaluator for tag name that ends with
      */
     public static final class TagEndsWith extends Evaluator {
-        private String tagName;
+        private final String tagName;
 
         public TagEndsWith(String tagName) {
             this.tagName = tagName;
@@ -64,7 +71,7 @@ public abstract class Evaluator {
 
         @Override
         public boolean matches(Element root, Element element) {
-            return (element.tagName().endsWith(tagName));
+            return (element.normalName().endsWith(tagName));
         }
 
         @Override
@@ -77,7 +84,7 @@ public abstract class Evaluator {
      * Evaluator for element id
      */
     public static final class Id extends Evaluator {
-        private String id;
+        private final String id;
 
         public Id(String id) {
             this.id = id;
@@ -99,7 +106,7 @@ public abstract class Evaluator {
      * Evaluator for element class
      */
     public static final class Class extends Evaluator {
-        private String className;
+        private final String className;
 
         public Class(String className) {
             this.className = className;
@@ -121,7 +128,7 @@ public abstract class Evaluator {
      * Evaluator for attribute name matching
      */
     public static final class Attribute extends Evaluator {
-        private String key;
+        private final String key;
 
         public Attribute(String key) {
             this.key = key;
@@ -143,18 +150,18 @@ public abstract class Evaluator {
      * Evaluator for attribute name prefix matching
      */
     public static final class AttributeStarting extends Evaluator {
-        private String keyPrefix;
+        private final String keyPrefix;
 
         public AttributeStarting(String keyPrefix) {
             Validate.notEmpty(keyPrefix);
-            this.keyPrefix = keyPrefix.toLowerCase();
+            this.keyPrefix = lowerCase(keyPrefix);
         }
 
         @Override
         public boolean matches(Element root, Element element) {
             List<org.jsoup.nodes.Attribute> values = element.attributes().asList();
             for (org.jsoup.nodes.Attribute attribute : values) {
-                if (attribute.getKey().toLowerCase().startsWith(keyPrefix))
+                if (lowerCase(attribute.getKey()).startsWith(keyPrefix))
                     return true;
             }
             return false;
@@ -212,12 +219,12 @@ public abstract class Evaluator {
      */
     public static final class AttributeWithValueStarting extends AttributeKeyPair {
         public AttributeWithValueStarting(String key, String value) {
-            super(key, value);
+            super(key, value, false);
         }
 
         @Override
         public boolean matches(Element root, Element element) {
-            return element.hasAttr(key) && element.attr(key).toLowerCase().startsWith(value); // value is lower case already
+            return element.hasAttr(key) && lowerCase(element.attr(key)).startsWith(value); // value is lower case already
         }
 
         @Override
@@ -232,12 +239,12 @@ public abstract class Evaluator {
      */
     public static final class AttributeWithValueEnding extends AttributeKeyPair {
         public AttributeWithValueEnding(String key, String value) {
-            super(key, value);
+            super(key, value, false);
         }
 
         @Override
         public boolean matches(Element root, Element element) {
-            return element.hasAttr(key) && element.attr(key).toLowerCase().endsWith(value); // value is lower case
+            return element.hasAttr(key) && lowerCase(element.attr(key)).endsWith(value); // value is lower case
         }
 
         @Override
@@ -257,7 +264,7 @@ public abstract class Evaluator {
 
         @Override
         public boolean matches(Element root, Element element) {
-            return element.hasAttr(key) && element.attr(key).toLowerCase().contains(value); // value is lower case
+            return element.hasAttr(key) && lowerCase(element.attr(key)).contains(value); // value is lower case
         }
 
         @Override
@@ -275,7 +282,7 @@ public abstract class Evaluator {
         Pattern pattern;
 
         public AttributeWithValueMatching(String key, Pattern pattern) {
-            this.key = key.trim().toLowerCase();
+            this.key = normalize(key);
             this.pattern = pattern;
         }
 
@@ -299,15 +306,21 @@ public abstract class Evaluator {
         String value;
 
         public AttributeKeyPair(String key, String value) {
+            this(key, value, true);
+        }
+
+        public AttributeKeyPair(String key, String value, boolean trimValue) {
             Validate.notEmpty(key);
             Validate.notEmpty(value);
 
-            this.key = key.trim().toLowerCase();
-            if (value.startsWith("\"") && value.endsWith("\"")
-                    || value.startsWith("'") && value.endsWith("'")) {
+            this.key = normalize(key);
+            boolean isStringLiteral = value.startsWith("'") && value.endsWith("'")
+                                        || value.startsWith("\"") && value.endsWith("\"");
+            if (isStringLiteral) {
                 value = value.substring(1, value.length()-1);
             }
-            this.value = value.trim().toLowerCase();
+
+            this.value = trimValue ? normalize(value) : normalize(value, isStringLiteral);
         }
     }
 
@@ -337,7 +350,7 @@ public abstract class Evaluator {
 
         @Override
         public boolean matches(Element root, Element element) {
-            return element.elementSiblingIndex() < index;
+            return root != element && element.elementSiblingIndex() < index;
         }
 
         @Override
@@ -386,7 +399,7 @@ public abstract class Evaluator {
         }
 
     }
-    
+
     /**
      * Evaluator for matching the last sibling (css :last-child)
      */
@@ -396,13 +409,13 @@ public abstract class Evaluator {
 			final Element p = element.parent();
 			return p != null && !(p instanceof Document) && element.elementSiblingIndex() == p.children().size()-1;
 		}
-    	
+
 		@Override
 		public String toString() {
 			return ":last-child";
 		}
     }
-    
+
     public static final class IsFirstOfType extends IsNthOfType {
 		public IsFirstOfType() {
 			super(0,1);
@@ -412,7 +425,7 @@ public abstract class Evaluator {
 			return ":first-of-type";
 		}
     }
-    
+
     public static final class IsLastOfType extends IsNthLastOfType {
 		public IsLastOfType() {
 			super(0,1);
@@ -423,10 +436,10 @@ public abstract class Evaluator {
 		}
     }
 
-    
+
     public static abstract class CssNthEvaluator extends Evaluator {
     	protected final int a, b;
-    	
+
     	public CssNthEvaluator(int a, int b) {
     		this.a = a;
     		this.b = b;
@@ -434,18 +447,18 @@ public abstract class Evaluator {
     	public CssNthEvaluator(int b) {
     		this(0,b);
     	}
-    	
+
     	@Override
     	public boolean matches(Element root, Element element) {
     		final Element p = element.parent();
     		if (p == null || (p instanceof Document)) return false;
-    		
+
     		final int pos = calculatePosition(root, element);
     		if (a == 0) return pos == b;
-    		
+
     		return (pos-b)*a >= 0 && (pos-b)%a==0;
     	}
-    	
+
 		@Override
 		public String toString() {
 			if (a == 0)
@@ -454,15 +467,15 @@ public abstract class Evaluator {
 				return String.format(":%s(%dn)",getPseudoClass(), a);
 			return String.format(":%s(%dn%+d)", getPseudoClass(),a, b);
 		}
-    	
+
 		protected abstract String getPseudoClass();
 		protected abstract int calculatePosition(Element root, Element element);
     }
-    
-    
+
+
     /**
      * css-compatible Evaluator for :eq (css :nth-child)
-     * 
+     *
      * @see IndexEquals
      */
     public static final class IsNthChild extends CssNthEvaluator {
@@ -475,15 +488,15 @@ public abstract class Evaluator {
 			return element.elementSiblingIndex()+1;
 		}
 
-		
+
 		protected String getPseudoClass() {
 			return "nth-child";
 		}
     }
-    
+
     /**
      * css pseudo class :nth-last-child)
-     * 
+     *
      * @see IndexEquals
      */
     public static final class IsNthLastChild extends CssNthEvaluator {
@@ -493,18 +506,20 @@ public abstract class Evaluator {
 
         @Override
         protected int calculatePosition(Element root, Element element) {
+    	    if (element.parent() == null)
+    	        return 0;
         	return element.parent().children().size() - element.elementSiblingIndex();
         }
-        
+
 		@Override
 		protected String getPseudoClass() {
 			return "nth-last-child";
 		}
     }
-    
+
     /**
      * css pseudo class nth-of-type
-     * 
+     *
      */
     public static class IsNthOfType extends CssNthEvaluator {
     	public IsNthOfType(int a, int b) {
@@ -513,6 +528,8 @@ public abstract class Evaluator {
 
 		protected int calculatePosition(Element root, Element element) {
 			int pos = 0;
+            if (element.parent() == null)
+                return 0;
         	Elements family = element.parent().children();
             for (Element el : family) {
                 if (el.tag().equals(element.tag())) pos++;
@@ -526,16 +543,18 @@ public abstract class Evaluator {
 			return "nth-of-type";
 		}
     }
-    
+
     public static class IsNthLastOfType extends CssNthEvaluator {
 
 		public IsNthLastOfType(int a, int b) {
 			super(a, b);
 		}
-		
+
 		@Override
 		protected int calculatePosition(Element root, Element element) {
 			int pos = 0;
+            if (element.parent() == null)
+                return 0;
         	Elements family = element.parent().children();
         	for (int i = element.elementSiblingIndex(); i < family.size(); i++) {
         		if (family.get(i).tag().equals(element.tag())) pos++;
@@ -558,13 +577,13 @@ public abstract class Evaluator {
     		final Element p = element.parent();
     		return p != null && !(p instanceof Document) && element.elementSiblingIndex() == 0;
     	}
-    	
+
     	@Override
     	public String toString() {
     		return ":first-child";
     	}
     }
-    
+
     /**
      * css3 pseudo-class :root
      * @see <a href="http://www.w3.org/TR/selectors/#root-pseudo">:root selector</a>
@@ -586,7 +605,7 @@ public abstract class Evaluator {
 		@Override
 		public boolean matches(Element root, Element element) {
 			final Element p = element.parent();
-			return p!=null && !(p instanceof Document) && element.siblingElements().size() == 0;
+			return p!=null && !(p instanceof Document) && element.siblingElements().isEmpty();
 		}
     	@Override
     	public String toString() {
@@ -599,7 +618,7 @@ public abstract class Evaluator {
 		public boolean matches(Element root, Element element) {
 			final Element p = element.parent();
 			if (p==null || p instanceof Document) return false;
-			
+
 			int pos = 0;
         	Elements family = p.children();
             for (Element el : family) {
@@ -645,15 +664,15 @@ public abstract class Evaluator {
      * Evaluator for matching Element (and its descendants) text
      */
     public static final class ContainsText extends Evaluator {
-        private String searchText;
+        private final String searchText;
 
         public ContainsText(String searchText) {
-            this.searchText = searchText.toLowerCase();
+            this.searchText = lowerCase(normaliseWhitespace(searchText));
         }
 
         @Override
         public boolean matches(Element root, Element element) {
-            return (element.text().toLowerCase().contains(searchText));
+            return lowerCase(element.text()).contains(searchText);
         }
 
         @Override
@@ -666,15 +685,15 @@ public abstract class Evaluator {
      * Evaluator for matching Element (and its descendants) data
      */
     public static final class ContainsData extends Evaluator {
-        private String searchText;
+        private final String searchText;
 
         public ContainsData(String searchText) {
-            this.searchText = searchText.toLowerCase();
+            this.searchText = lowerCase(searchText);
         }
 
         @Override
         public boolean matches(Element root, Element element) {
-            return (element.data().toLowerCase().contains(searchText));
+            return lowerCase(element.data()).contains(searchText); // not whitespace normalized
         }
 
         @Override
@@ -687,15 +706,15 @@ public abstract class Evaluator {
      * Evaluator for matching Element's own text
      */
     public static final class ContainsOwnText extends Evaluator {
-        private String searchText;
+        private final String searchText;
 
         public ContainsOwnText(String searchText) {
-            this.searchText = searchText.toLowerCase();
+            this.searchText = lowerCase(normaliseWhitespace(searchText));
         }
 
         @Override
         public boolean matches(Element root, Element element) {
-            return (element.ownText().toLowerCase().contains(searchText));
+            return lowerCase(element.ownText()).contains(searchText);
         }
 
         @Override
@@ -708,7 +727,7 @@ public abstract class Evaluator {
      * Evaluator for matching Element (and its descendants) text with regex
      */
     public static final class Matches extends Evaluator {
-        private Pattern pattern;
+        private final Pattern pattern;
 
         public Matches(Pattern pattern) {
             this.pattern = pattern;
@@ -730,7 +749,7 @@ public abstract class Evaluator {
      * Evaluator for matching Element's own text with regex
      */
     public static final class MatchesOwn extends Evaluator {
-        private Pattern pattern;
+        private final Pattern pattern;
 
         public MatchesOwn(Pattern pattern) {
             this.pattern = pattern;
@@ -747,4 +766,27 @@ public abstract class Evaluator {
             return String.format(":matchesOwn(%s)", pattern);
         }
     }
+
+    public static final class MatchText extends Evaluator {
+
+        @Override
+        public boolean matches(Element root, Element element) {
+            if (element instanceof PseudoTextElement)
+                return true;
+
+            List<TextNode> textNodes = element.textNodes();
+            for (TextNode textNode : textNodes) {
+                PseudoTextElement pel = new PseudoTextElement(
+                    org.jsoup.parser.Tag.valueOf(element.tagName()), element.baseUri(), element.attributes());
+                textNode.replaceWith(pel);
+                pel.appendChild(textNode);
+            }
+            return false;
+        }
+
+        @Override
+        public String toString() {
+            return ":matchText";
+        }
+    }
 }
