@@ -5,21 +5,17 @@ import org.jsoup.TextUtil;
 import org.jsoup.integration.ParseTest;
 import org.jsoup.nodes.Document.OutputSettings;
 import org.jsoup.nodes.Document.OutputSettings.Syntax;
-import org.junit.Ignore;
-import org.junit.Test;
+import org.jsoup.parser.ParseSettings;
+import org.jsoup.parser.Parser;
+import org.jsoup.select.Elements;
+import org.junit.jupiter.api.Disabled;
+import org.junit.jupiter.api.Test;
 
-import java.io.ByteArrayInputStream;
-import java.io.File;
-import java.io.IOException;
-import java.io.InputStream;
-import java.io.StringWriter;
+import java.io.*;
 import java.nio.charset.Charset;
+import java.nio.charset.StandardCharsets;
 
-import static org.junit.Assert.assertEquals;
-import static org.junit.Assert.assertFalse;
-import static org.junit.Assert.assertNotNull;
-import static org.junit.Assert.assertNull;
-import static org.junit.Assert.assertTrue;
+import static org.junit.jupiter.api.Assertions.*;
 
 /**
  Tests for Document.
@@ -28,8 +24,8 @@ import static org.junit.Assert.assertTrue;
 public class DocumentTest {
     private static final String charsetUtf8 = "UTF-8";
     private static final String charsetIso8859 = "ISO-8859-1";
-    
-    
+
+
     @Test public void setTextPreservesDocumentStructure() {
         Document doc = Jsoup.parse("<p>Hello</p>");
         doc.text("Replaced");
@@ -37,16 +33,16 @@ public class DocumentTest {
         assertEquals("Replaced", doc.body().text());
         assertEquals(1, doc.select("head").size());
     }
-    
+
     @Test public void testTitles() {
         Document noTitle = Jsoup.parse("<p>Hello</p>");
         Document withTitle = Jsoup.parse("<title>First</title><title>Ignore</title><p>Hello</p>");
-        
+
         assertEquals("", noTitle.title());
         noTitle.title("Hello");
         assertEquals("Hello", noTitle.title());
         assertEquals("Hello", noTitle.select("title").first().text());
-        
+
         assertEquals("First", withTitle.title());
         withTitle.title("Hello");
         assertEquals("Hello", withTitle.title());
@@ -81,6 +77,34 @@ public class DocumentTest {
         assertEquals("<html><head><script>one</script><noscript>&lt;p&gt;two</noscript></head><body><p>three</p><p>four</p></body></html>", TextUtil.stripNewlines(doc.html()));
     }
 
+    @Test public void accessorsWillNormalizeStructure() {
+        Document doc = new Document("");
+        assertEquals("", doc.html());
+
+        Element body = doc.body();
+        assertEquals("body", body.tagName());
+        Element head = doc.head();
+        assertEquals("head", head.tagName());
+        assertEquals("<html><head></head><body></body></html>", TextUtil.stripNewlines(doc.html()));
+    }
+
+    @Test public void accessorsAreCaseInsensitive() {
+        Parser parser = Parser.htmlParser().settings(ParseSettings.preserveCase);
+        Document doc = parser.parseInput("<!DOCTYPE html><HTML><HEAD><TITLE>SHOUTY</TITLE></HEAD><BODY>HELLO</BODY></HTML>", "");
+
+        Element body = doc.body();
+        assertEquals("BODY", body.tagName());
+        assertEquals("body", body.normalName());
+        Element head = doc.head();
+        assertEquals("HEAD", head.tagName());
+        assertEquals("body", body.normalName());
+
+        Element root = doc.selectFirst("html");
+        assertEquals("HTML", root.tagName());
+        assertEquals("html", root.normalName());
+        assertEquals("SHOUTY", doc.title());
+    }
+
     @Test public void testClone() {
         Document doc = Jsoup.parse("<title>Hello</title> <p>One<p>Two");
         Document clone = doc.clone();
@@ -100,15 +124,15 @@ public class DocumentTest {
         assertEquals("<!doctype html><html><head><title>Doctype test</title></head><body></body></html>",
                 TextUtil.stripNewlines(clone.html()));
     }
-    
+
     @Test public void testLocation() throws IOException {
-    	File in = new ParseTest().getFile("/htmltests/yahoo-jp.html");
+    	File in = ParseTest.getFile("/htmltests/yahoo-jp.html.gz");
         Document doc = Jsoup.parse(in, "UTF-8", "http://www.yahoo.co.jp/index.html");
         String location = doc.location();
         String baseUri = doc.baseUri();
         assertEquals("http://www.yahoo.co.jp/index.html",location);
         assertEquals("http://www.yahoo.co.jp/_ylh=X3oDMTB0NWxnaGxsBF9TAzIwNzcyOTYyNjUEdGlkAzEyBHRtcGwDZ2Ex/",baseUri);
-        in = new ParseTest().getFile("/htmltests/nyt-article-1.html");
+        in = ParseTest.getFile("/htmltests/nyt-article-1.html.gz");
         doc = Jsoup.parse(in, null, "http://www.nytimes.com/2010/07/26/business/global/26bp.html?hp");
         location = doc.location();
         baseUri = doc.baseUri();
@@ -116,6 +140,11 @@ public class DocumentTest {
         assertEquals("http://www.nytimes.com/2010/07/26/business/global/26bp.html?hp",baseUri);
     }
 
+    @Test public void testLocationFromString() {
+        Document doc = Jsoup.parse("<p>Hello");
+        assertEquals("", doc.location());
+    }
+
     @Test public void testHtmlAndXmlSyntax() {
         String h = "<!DOCTYPE html><body><img async checked='checked' src='&<>\"'>&lt;&gt;&amp;&quot;<foo />bar";
         Document doc = Jsoup.parse(h);
@@ -125,8 +154,7 @@ public class DocumentTest {
                 "<html>\n" +
                 " <head></head>\n" +
                 " <body>\n" +
-                "  <img async checked src=\"&amp;<>&quot;\">&lt;&gt;&amp;\"\n" +
-                "  <foo />bar\n" +
+                "  <img async checked src=\"&amp;<>&quot;\">&lt;&gt;&amp;\"<foo />bar\n" +
                 " </body>\n" +
                 "</html>", doc.html());
 
@@ -135,8 +163,7 @@ public class DocumentTest {
                 "<html>\n" +
                 " <head></head>\n" +
                 " <body>\n" +
-                "  <img async=\"\" checked=\"checked\" src=\"&amp;<>&quot;\" />&lt;&gt;&amp;\"\n" +
-                "  <foo />bar\n" +
+                "  <img async=\"\" checked=\"checked\" src=\"&amp;&lt;>&quot;\" />&lt;&gt;&amp;\"<foo />bar\n" +
                 " </body>\n" +
                 "</html>", doc.html());
     }
@@ -145,42 +172,46 @@ public class DocumentTest {
         Document doc = Jsoup.parse("x");
         assertEquals(Syntax.html, doc.outputSettings().syntax());
     }
-    
+
     @Test public void testHtmlAppendable() {
     	String htmlContent = "<html><head><title>Hello</title></head><body><p>One</p><p>Two</p></body></html>";
     	Document document = Jsoup.parse(htmlContent);
     	OutputSettings outputSettings = new OutputSettings();
-    	
+
     	outputSettings.prettyPrint(false);
     	document.outputSettings(outputSettings);
     	assertEquals(htmlContent, document.html(new StringWriter()).toString());
     }
 
-    // Ignored since this test can take awhile to run.
-    @Ignore
     @Test public void testOverflowClone() {
-        StringBuilder builder = new StringBuilder();
+        StringBuilder sb = new StringBuilder();
+        sb.append("<head><base href='https://jsoup.org/'>");
         for (int i = 0; i < 100000; i++) {
-            builder.insert(0, "<i>");
-            builder.append("</i>");
+            sb.append("<div>");
         }
+        sb.append("<p>Hello <a href='/example.html'>there</a>");
 
-        Document doc = Jsoup.parse(builder.toString());
-        doc.clone();
+        Document doc = Jsoup.parse(sb.toString());
+
+        String expectedLink = "https://jsoup.org/example.html";
+        assertEquals(expectedLink, doc.selectFirst("a").attr("abs:href"));
+        Document clone = doc.clone();
+        doc.hasSameValue(clone);
+        assertEquals(expectedLink, clone.selectFirst("a").attr("abs:href"));
     }
 
-    @Test public void DocumentsWithSameContentAreEqual() throws Exception {
+    @Test public void DocumentsWithSameContentAreEqual() {
         Document docA = Jsoup.parse("<div/>One");
         Document docB = Jsoup.parse("<div/>One");
         Document docC = Jsoup.parse("<div/>Two");
 
-        assertFalse(docA.equals(docB));
-        assertTrue(docA.equals(docA));
+        assertNotEquals(docA, docB);
+        assertEquals(docA, docA);
         assertEquals(docA.hashCode(), docA.hashCode());
-        assertFalse(docA.hashCode() == docC.hashCode());
+        assertNotEquals(docA.hashCode(), docC.hashCode());
     }
 
-    @Test public void DocumentsWithSameContentAreVerifialbe() throws Exception {
+    @Test public void DocumentsWithSameContentAreVerifiable() {
         Document docA = Jsoup.parse("<div/>One");
         Document docB = Jsoup.parse("<div/>One");
         Document docC = Jsoup.parse("<div/>Two");
@@ -188,13 +219,13 @@ public class DocumentTest {
         assertTrue(docA.hasSameValue(docB));
         assertFalse(docA.hasSameValue(docC));
     }
-    
+
     @Test
     public void testMetaCharsetUpdateUtf8() {
         final Document doc = createHtmlDocument("changeThis");
         doc.updateMetaCharsetElement(true);
         doc.charset(Charset.forName(charsetUtf8));
-        
+
         final String htmlCharsetUTF8 = "<html>\n" +
                                         " <head>\n" +
                                         "  <meta charset=\"" + charsetUtf8 + "\">\n" +
@@ -202,19 +233,19 @@ public class DocumentTest {
                                         " <body></body>\n" +
                                         "</html>";
         assertEquals(htmlCharsetUTF8, doc.toString());
-        
+
         Element selectedElement = doc.select("meta[charset]").first();
         assertEquals(charsetUtf8, doc.charset().name());
         assertEquals(charsetUtf8, selectedElement.attr("charset"));
         assertEquals(doc.charset(), doc.outputSettings().charset());
     }
-    
+
     @Test
     public void testMetaCharsetUpdateIso8859() {
         final Document doc = createHtmlDocument("changeThis");
         doc.updateMetaCharsetElement(true);
         doc.charset(Charset.forName(charsetIso8859));
-        
+
         final String htmlCharsetISO = "<html>\n" +
                                         " <head>\n" +
                                         "  <meta charset=\"" + charsetIso8859 + "\">\n" +
@@ -222,34 +253,34 @@ public class DocumentTest {
                                         " <body></body>\n" +
                                         "</html>";
         assertEquals(htmlCharsetISO, doc.toString());
-        
+
         Element selectedElement = doc.select("meta[charset]").first();
         assertEquals(charsetIso8859, doc.charset().name());
         assertEquals(charsetIso8859, selectedElement.attr("charset"));
         assertEquals(doc.charset(), doc.outputSettings().charset());
     }
-    
+
     @Test
     public void testMetaCharsetUpdateNoCharset() {
         final Document docNoCharset = Document.createShell("");
         docNoCharset.updateMetaCharsetElement(true);
         docNoCharset.charset(Charset.forName(charsetUtf8));
-        
+
         assertEquals(charsetUtf8, docNoCharset.select("meta[charset]").first().attr("charset"));
-        
+
         final String htmlCharsetUTF8 = "<html>\n" +
                                         " <head>\n" +
                                         "  <meta charset=\"" + charsetUtf8 + "\">\n" +
                                         " </head>\n" +
                                         " <body></body>\n" +
                                         "</html>";
-        assertEquals(htmlCharsetUTF8, docNoCharset.toString()); 
+        assertEquals(htmlCharsetUTF8, docNoCharset.toString());
     }
-    
+
     @Test
     public void testMetaCharsetUpdateDisabled() {
         final Document docDisabled = Document.createShell("");
-        
+
         final String htmlNoCharset = "<html>\n" +
                                         " <head></head>\n" +
                                         " <body></body>\n" +
@@ -257,11 +288,11 @@ public class DocumentTest {
         assertEquals(htmlNoCharset, docDisabled.toString());
         assertNull(docDisabled.select("meta[charset]").first());
     }
-    
+
     @Test
     public void testMetaCharsetUpdateDisabledNoChanges() {
         final Document doc = createHtmlDocument("dontTouch");
-        
+
         final String htmlCharset = "<html>\n" +
                                     " <head>\n" +
                                     "  <meta charset=\"dontTouch\">\n" +
@@ -270,48 +301,48 @@ public class DocumentTest {
                                     " <body></body>\n" +
                                     "</html>";
         assertEquals(htmlCharset, doc.toString());
-        
+
         Element selectedElement = doc.select("meta[charset]").first();
         assertNotNull(selectedElement);
         assertEquals("dontTouch", selectedElement.attr("charset"));
-        
+
         selectedElement = doc.select("meta[name=charset]").first();
         assertNotNull(selectedElement);
         assertEquals("dontTouch", selectedElement.attr("content"));
     }
-    
+
     @Test
     public void testMetaCharsetUpdateEnabledAfterCharsetChange() {
         final Document doc = createHtmlDocument("dontTouch");
         doc.charset(Charset.forName(charsetUtf8));
-        
+
         Element selectedElement = doc.select("meta[charset]").first();
         assertEquals(charsetUtf8, selectedElement.attr("charset"));
         assertTrue(doc.select("meta[name=charset]").isEmpty());
     }
-            
+
     @Test
     public void testMetaCharsetUpdateCleanup() {
         final Document doc = createHtmlDocument("dontTouch");
         doc.updateMetaCharsetElement(true);
         doc.charset(Charset.forName(charsetUtf8));
-        
+
         final String htmlCharsetUTF8 = "<html>\n" +
                                         " <head>\n" +
                                         "  <meta charset=\"" + charsetUtf8 + "\">\n" +
                                         " </head>\n" +
                                         " <body></body>\n" +
                                         "</html>";
-        
+
         assertEquals(htmlCharsetUTF8, doc.toString());
     }
-    
+
     @Test
     public void testMetaCharsetUpdateXmlUtf8() {
         final Document doc = createXmlDocument("1.0", "changeThis", true);
         doc.updateMetaCharsetElement(true);
         doc.charset(Charset.forName(charsetUtf8));
-        
+
         final String xmlCharsetUTF8 = "<?xml version=\"1.0\" encoding=\"" + charsetUtf8 + "\"?>\n" +
                                         "<root>\n" +
                                         " node\n" +
@@ -323,45 +354,45 @@ public class DocumentTest {
         assertEquals(charsetUtf8, selectedNode.attr("encoding"));
         assertEquals(doc.charset(), doc.outputSettings().charset());
     }
-    
+
     @Test
     public void testMetaCharsetUpdateXmlIso8859() {
         final Document doc = createXmlDocument("1.0", "changeThis", true);
         doc.updateMetaCharsetElement(true);
         doc.charset(Charset.forName(charsetIso8859));
-        
+
         final String xmlCharsetISO = "<?xml version=\"1.0\" encoding=\"" + charsetIso8859 + "\"?>\n" +
                                         "<root>\n" +
                                         " node\n" +
                                         "</root>";
         assertEquals(xmlCharsetISO, doc.toString());
-        
+
         XmlDeclaration selectedNode = (XmlDeclaration) doc.childNode(0);
         assertEquals(charsetIso8859, doc.charset().name());
         assertEquals(charsetIso8859, selectedNode.attr("encoding"));
         assertEquals(doc.charset(), doc.outputSettings().charset());
     }
-    
+
     @Test
     public void testMetaCharsetUpdateXmlNoCharset() {
         final Document doc = createXmlDocument("1.0", "none", false);
         doc.updateMetaCharsetElement(true);
         doc.charset(Charset.forName(charsetUtf8));
-        
+
         final String xmlCharsetUTF8 = "<?xml version=\"1.0\" encoding=\"" + charsetUtf8 + "\"?>\n" +
                                         "<root>\n" +
                                         " node\n" +
                                         "</root>";
         assertEquals(xmlCharsetUTF8, doc.toString());
-        
+
         XmlDeclaration selectedNode = (XmlDeclaration) doc.childNode(0);
         assertEquals(charsetUtf8, selectedNode.attr("encoding"));
     }
-    
+
     @Test
     public void testMetaCharsetUpdateXmlDisabled() {
         final Document doc = createXmlDocument("none", "none", false);
-        
+
         final String xmlNoCharset = "<root>\n" +
                                     " node\n" +
                                     "</root>";
@@ -371,44 +402,44 @@ public class DocumentTest {
     @Test
     public void testMetaCharsetUpdateXmlDisabledNoChanges() {
         final Document doc = createXmlDocument("dontTouch", "dontTouch", true);
-        
+
         final String xmlCharset = "<?xml version=\"dontTouch\" encoding=\"dontTouch\"?>\n" +
                                     "<root>\n" +
                                     " node\n" +
                                     "</root>";
         assertEquals(xmlCharset, doc.toString());
-        
+
         XmlDeclaration selectedNode = (XmlDeclaration) doc.childNode(0);
         assertEquals("dontTouch", selectedNode.attr("encoding"));
         assertEquals("dontTouch", selectedNode.attr("version"));
     }
-    
+
     @Test
     public void testMetaCharsetUpdatedDisabledPerDefault() {
         final Document doc = createHtmlDocument("none");
         assertFalse(doc.updateMetaCharsetElement());
     }
-    
+
     private Document createHtmlDocument(String charset) {
         final Document doc = Document.createShell("");
         doc.head().appendElement("meta").attr("charset", charset);
         doc.head().appendElement("meta").attr("name", "charset").attr("content", charset);
-        
+
         return doc;
     }
-    
+
     private Document createXmlDocument(String version, String charset, boolean addDecl) {
         final Document doc = new Document("");
         doc.appendElement("root").text("node");
         doc.outputSettings().syntax(Syntax.xml);
-        
-        if( addDecl == true ) {
-            XmlDeclaration decl = new XmlDeclaration("xml", "", false);
+
+        if(addDecl) {
+            XmlDeclaration decl = new XmlDeclaration("xml", false);
             decl.attr("version", version);
             decl.attr("encoding", charset);
             doc.prependChild(decl);
         }
-        
+
         return doc;
     }
 
@@ -423,15 +454,76 @@ public class DocumentTest {
                         +     "before&nbsp;after"
                         +   "</body>"
                         + "</html>";
-        InputStream is = new ByteArrayInputStream(input.getBytes(Charset.forName("ASCII")));
+        InputStream is = new ByteArrayInputStream(input.getBytes(StandardCharsets.US_ASCII));
 
         Document doc = Jsoup.parse(is, null, "http://example.com");
         doc.outputSettings().escapeMode(Entities.EscapeMode.xhtml);
 
         String output = new String(doc.html().getBytes(doc.outputSettings().charset()), doc.outputSettings().charset());
 
-        assertFalse("Should not have contained a '?'.", output.contains("?"));
-        assertTrue("Should have contained a '&#xa0;' or a '&nbsp;'.",
-                output.contains("&#xa0;") || output.contains("&nbsp;"));
+        assertFalse(output.contains("?"), "Should not have contained a '?'.");
+        assertTrue(output.contains("&#xa0;") || output.contains("&nbsp;"),
+                "Should have contained a '&#xa0;' or a '&nbsp;'.");
+    }
+
+    @Test public void parseAndHtmlOnDifferentThreads() throws InterruptedException {
+        String html = "<p>Alrighty then it's not \uD83D\uDCA9. <span>Next</span></p>"; // 💩
+        String asci = "<p>Alrighty then it's not &#x1f4a9;. <span>Next</span></p>";
+
+        final Document doc = Jsoup.parse(html);
+        final String[] out = new String[1];
+        final Elements p = doc.select("p");
+        assertEquals(html, p.outerHtml());
+
+        Thread thread = new Thread(() -> {
+            out[0] = p.outerHtml();
+            doc.outputSettings().charset(StandardCharsets.US_ASCII);
+        });
+        thread.start();
+        thread.join();
+
+        assertEquals(html, out[0]);
+        assertEquals(StandardCharsets.US_ASCII, doc.outputSettings().charset());
+        assertEquals(asci, p.outerHtml());
+    }
+
+    @Test public void testDocumentTypeGet() {
+        String html = "\n\n<!-- comment -->  <!doctype html><p>One</p>";
+        Document doc = Jsoup.parse(html);
+        DocumentType documentType = doc.documentType();
+        assertNotNull(documentType);
+        assertEquals("html", documentType.name());
+    }
+
+    @Test public void framesetSupportsBodyMethod() {
+        String html = "<html><head><title>Frame Test</title></head><frameset id=id><frame src=foo.html></frameset>";
+        Document doc = Jsoup.parse(html);
+        Element head = doc.head();
+        assertNotNull(head);
+        assertEquals("Frame Test", doc.title());
+
+        // Frameset docs per html5 spec have no body element - but instead a frameset elelemt
+        assertNull(doc.selectFirst("body"));
+        Element frameset = doc.selectFirst("frameset");
+        assertNotNull(frameset);
+
+        // the body() method returns body or frameset and does not otherwise modify the document
+        // doing it in body() vs parse keeps the html close to original for round-trip option
+        Element body = doc.body();
+        assertNotNull(body);
+        assertSame(frameset, body);
+        assertEquals("frame", body.child(0).tagName());
+
+        assertNull(doc.selectFirst("body")); // did not vivify a body element
+
+        String expected = "<html>\n" +
+            " <head>\n" +
+            "  <title>Frame Test</title>\n" +
+            " </head>\n" +
+            " <frameset id=\"id\">\n" +
+            "  <frame src=\"foo.html\">\n" +
+            " </frameset>\n" +
+            "</html>";
+        assertEquals(expected, doc.html());
     }
 }
