@@ -2,6 +2,7 @@ package org.jsoup.nodes;
 
 import org.jsoup.Jsoup;
 import org.jsoup.TextUtil;
+import org.jsoup.parser.ParseSettings;
 import org.jsoup.parser.Parser;
 import org.jsoup.parser.Tag;
 import org.jsoup.select.Elements;
@@ -161,6 +162,16 @@ public class ElementTest {
         assertEquals("Hello there", doc.text());
     }
 
+    @Test
+    public void testBrHasSpaceCaseSensitive() {
+        Document doc = Jsoup.parse("<p>Hello<br>there<BR>now</p>", Parser.htmlParser().settings(ParseSettings.preserveCase));
+        assertEquals("Hello there now", doc.text());
+        assertEquals("Hello there now", doc.select("p").first().ownText());
+
+        doc = Jsoup.parse("<p>Hello <br> there <BR> now</p>");
+        assertEquals("Hello there now", doc.text());
+    }
+
     @Test
     public void testWholeText() {
         Document doc = Jsoup.parse("<p> Hello\nthere &nbsp;  </p>");
